import os
import subprocess
import sys
from ailab.log import logger


def install_requiremet(basedir: str):
    basereqs = os.path.join(basedir, "requirements.txt")
    if not os.path.exists(basereqs):
        return
    cmd = "pip install -r  " + basereqs
    print(f"executing {cmd}")

    subprocess.call(cmd, shell=True)


from typing import Dict, List, Optional
import os, json, torch
from transformers.trainer import TRAINER_STATE_NAME

VALUE_HEAD_FILE_NAME = "value_head.bin"


def get_state_dict(model: torch.nn.Module) -> Dict[str, torch.Tensor]:  # get state dict containing trainable parameters
    state_dict = model.state_dict()
    filtered_state_dict = {}

    for k, v in model.named_parameters():
        if v.requires_grad:
            filtered_state_dict[k] = state_dict[k].cpu().clone().detach()
    return filtered_state_dict


def load_trainable_params(model: torch.nn.Module, checkpoint_dir: os.PathLike) -> bool:
    from peft.utils.other import WEIGHTS_NAME
    weights_file = os.path.join(checkpoint_dir, WEIGHTS_NAME)
    if not os.path.exists(weights_file):
        logger.info("Provided path ({}) does not contain pre-trained weights.".format(checkpoint_dir))
        return False
    model_state_dict = torch.load(weights_file, map_location="cpu")
    model.load_state_dict(model_state_dict, strict=False)  # skip missing keys
    return True

def print_trainable_params(model: torch.nn.Module) -> None:
    trainable_params, all_param = 0, 0
    for param in model.parameters():
        num_params = param.numel()
        # if using DS Zero 3 and the weights are initialized empty
        if num_params == 0 and hasattr(param, "ds_numel"):
            num_params = param.ds_numel
        all_param += num_params
        if param.requires_grad:
            trainable_params += num_params
    logger.info("trainable params: {:d} || all params: {:d} || trainable%: {:.4f}".format(
                trainable_params, all_param, 100 * trainable_params / all_param))

def load_valuehead_params(model: torch.nn.Module, checkpoint_dir: os.PathLike) -> bool:
    valuehead_file = os.path.join(checkpoint_dir, VALUE_HEAD_FILE_NAME)
    if not os.path.exists(valuehead_file):
        logger.info("Provided path ({}) does not contain valuehead weights.".format(checkpoint_dir))
        return False
    valuehead_state_dict = torch.load(valuehead_file, map_location="cpu")
    model.register_buffer("reward_head_weight", valuehead_state_dict["summary.weight"])
    model.register_buffer("reward_head_bias", valuehead_state_dict["summary.bias"])
    model.register_buffer("default_head_weight", torch.zeros_like(valuehead_state_dict["summary.weight"]))
    model.register_buffer("default_head_bias", torch.zeros_like(valuehead_state_dict["summary.bias"]))
    return True


def smooth(scalars: List[float], weight: Optional[float] = 0.9) -> List[float]:
    r"""
    EMA implementation according to TensorBoard.
    """
    last = scalars[0]
    smoothed = list()
    for next_val in scalars:
        smoothed_val = last * weight + (1 - weight) * next_val
        smoothed.append(smoothed_val)
        last = smoothed_val
    return smoothed


def plot_loss(save_dictionary: os.PathLike, keys: Optional[List[str]] = ["loss"]) -> None:
    import matplotlib.pyplot as plt
    with open(os.path.join(save_dictionary, TRAINER_STATE_NAME), "r", encoding="utf-8") as f:
        data = json.load(f)

    for key in keys:
        steps, metrics = [], []
        for i in range(len(data["log_history"])):
            if key in data["log_history"][i]:
                steps.append(data["log_history"][i]["step"])
                metrics.append(data["log_history"][i][key])

        if len(metrics) == 0:
            logger.info(f"No metric {key} to plot.")
            continue

        plt.figure()
        plt.plot(steps, metrics, alpha=0.4, label="original")
        plt.plot(steps, smooth(metrics), label="smoothed")
        plt.title("training {} of {}".format(key, save_dictionary))
        plt.xlabel("step")
        plt.ylabel(key)
        plt.legend()
        plt.savefig(os.path.join(save_dictionary, "training_{}.png".format(key)), format="png", dpi=100)
        logger.info("Figure saved:", os.path.join(save_dictionary, "training_{}.png".format(key)))


import os
import zipfile
import shutil


def create_zip_and_delete_folder(folder_path, zip_path):
    try:
        files_to_compress = ['adapter_model.bin', 'adapter_config.json', 'pytorch_lora_weights.safetensors', 'adapter_model.safetensors']  # 要压缩的文件列表

        # 创建一个 ZIP 文件并添加文件到其中
        with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for file_name in files_to_compress:
                file_path = os.path.join(folder_path, file_name)
                if os.path.exists(file_path):
                    zipf.write(file_path, os.path.basename(file_path))

        # 删除 checkpoint-* 文件
        for dirpath, dirnames, filenames in os.walk(folder_path):
            for dirname in dirnames:
                if dirname.startswith("checkpoint-"):
                    dir_to_delete = os.path.join(dirpath, dirname)
                    print(f"Deleting directory: {dir_to_delete}")
                    shutil.rmtree(dir_to_delete)
        
        # 删除 runs 文件夹
        runs_folder = os.path.join(folder_path, "runs")
        if os.path.exists(runs_folder):
            shutil.rmtree(runs_folder)

        print(f"文件夹 '{folder_path}' 已打包为 '{zip_path}' 并删除其他文件")
    except Exception as e:
        print(f"发生错误：{e}")

def is_deepspeed_available(**kwargs):
    finetune_type = kwargs['model_args'].get('finetune_type', 'lora')
    loraplus_lr_ratio = kwargs['model_args'].get('loraplus_lr_ratio', 0)
    use_galore = kwargs['model_args'].get('use_galore', False)
    if finetune_type == 'full' and use_galore:
        print(f'galore cannot use deepspeed')
        return False
    if finetune_type == 'lora' and loraplus_lr_ratio > 0:
        print(f'loraplus cannot use deepspeed')
        return False
    return True

from typing import Union,Callable
import math
from datasets import Dataset,IterableDataset
from transformers import PreTrainedModel,Seq2SeqTrainingArguments,Trainer
from transformers.utils.versions import require_version
from transformers.optimization import get_scheduler

class DummyOptimizer(torch.optim.Optimizer):
    def __init__(self, *args, **kwargs):
        dummy_tensor = torch.randn(1, 1)
        super().__init__([dummy_tensor], {"lr": 1e-3})

    def zero_grad(self, set_to_none: bool = True) -> None:
        pass

    def step(self, closure: Optional[Callable[[], float]] = None) -> Optional[float]:
        pass

def _get_decay_parameter_names(model: "PreTrainedModel") -> List[str]:
    r"""
    Returns a list of names of parameters with weight decay. (weights in non-layernorm layers)
    """
    from transformers.pytorch_utils import ALL_LAYERNORM_LAYERS
    from transformers.trainer_pt_utils import get_parameter_names
    decay_parameters = get_parameter_names(model, ALL_LAYERNORM_LAYERS)
    decay_parameters = [name for name in decay_parameters if "bias" not in name]
    return decay_parameters

def create_loraplus_optimizer(
    model: "PreTrainedModel",
    training_args: "Seq2SeqTrainingArguments",
    **kwargs
) -> "torch.optim.Optimizer":
    loraplus_lr_ratio = kwargs['model_args'].get('loraplus_lr_ratio', 0)
    loraplus_lr = training_args.learning_rate * loraplus_lr_ratio
    decay_args = {"weight_decay": training_args.weight_decay}

    decay_param_names = _get_decay_parameter_names(model)
    param_dict: Dict[str, List["torch.nn.Parameter"]] = {
        "lora_a": [],
        "lora_b": [],
        "lora_b_nodecay": [],
        "embedding": [],
    }
    for name, param in model.named_parameters():
        if param.requires_grad:
            if "lora_embedding_B" in name:
                param_dict["embedding"].append(param)
            elif "lora_B" in name or param.ndim == 1:
                if name in decay_param_names:
                    param_dict["lora_b"].append(param)
                else:
                    param_dict["lora_b_nodecay"].append(param)
            else:
                param_dict["lora_a"].append(param)

    optim_class, optim_kwargs = Trainer.get_optimizer_cls_and_kwargs(training_args)
    loraplus_lr_embedding = 1e-6
    param_groups = [
        dict(params=param_dict["lora_a"], **decay_args),
        dict(params=param_dict["lora_b"], lr=loraplus_lr, **decay_args),
        dict(params=param_dict["lora_b_nodecay"], lr=loraplus_lr),
        dict(params=param_dict["embedding"], lr=loraplus_lr_embedding, **decay_args),
    ]
    optimizer = optim_class(param_groups, **optim_kwargs)
    logger.info("Using LoRA+ optimizer with loraplus lr ratio {:.2f}.".format(loraplus_lr_ratio))
    return optimizer

def create_galore_optimzer(
    model: "PreTrainedModel",
    dataset: Union["Dataset", "IterableDataset"],
    training_args: "Seq2SeqTrainingArguments",
) -> Optional["torch.optim.Optimizer"]:

    require_version("galore_torch", "To fix: pip install git+https://github.com/hiyouga/GaLore.git")
    galore_params: List[torch.nn.Parameter] = []
    galore_targets = ['mlp','self_attn']

    for name, module in model.named_modules():
        if isinstance(module, torch.nn.Linear) and any(target in name for target in galore_targets):
            for param in module.parameters():
                if param.requires_grad and len(param.shape) > 1:
                    galore_params.append(param)

    id_galore_params = {id(param) for param in galore_params}
    trainable_params = filter(lambda param: param.requires_grad, model.parameters())
    non_galore_params = [param for param in trainable_params if id(param) not in id_galore_params]
    
    from galore_torch import GaLoreAdamW
    optim_class = GaLoreAdamW
    optim_kwargs = {
        "lr": training_args.learning_rate,
        "eps": training_args.adam_epsilon,
        "betas": (training_args.adam_beta1, training_args.adam_beta2),
        "weight_decay": training_args.weight_decay,
    }

    galore_rank = 128
    galore_update_interval = 200
    galore_scale = 0.25
    galore_proj_type = "std"
    galore_layerwise = False
    galore_kwargs = {
        "rank": galore_rank,
        "update_proj_gap": galore_update_interval,
        "scale": galore_scale,
        "proj_type": galore_proj_type,
    }

    if galore_layerwise:
        if training_args.gradient_accumulation_steps != 1:
            raise ValueError("Per-layer GaLore does not support gradient accumulation.")

        if training_args.max_steps > 0:
            num_training_steps = training_args.max_steps
        else:
            total_train_batch_size = training_args.per_device_train_batch_size * training_args.world_size
            num_training_steps = training_args.num_train_epochs * math.ceil(len(dataset) / total_train_batch_size)

        optimizer_dict: Dict["torch.Tensor", "torch.optim.Optimizer"] = {}
        for param in non_galore_params:
            param_groups = [dict(params=[param])]
            optimizer_dict[param] = optim_class(param_groups, **optim_kwargs)
        for param in galore_params:
            param_groups = [dict(params=[param], **galore_kwargs)]
            optimizer_dict[param] = optim_class(param_groups, **optim_kwargs)

        scheduler_dict: Dict["torch.Tensor", "torch.optim.lr_scheduler.LRScheduler"] = {}
        for param in non_galore_params + galore_params:
            scheduler_dict[param] = get_scheduler(
                training_args.lr_scheduler_type,
                optimizer=optimizer_dict[param],
                num_warmup_steps=training_args.get_warmup_steps(num_training_steps) * 2,
                num_training_steps=num_training_steps * 2,
            )

        def optimizer_hook(param: "torch.Tensor"):
            if param.grad is not None:
                optimizer_dict[param].step()
                optimizer_dict[param].zero_grad()
                scheduler_dict[param].step()

        for param in non_galore_params + galore_params:
            param.register_post_accumulate_grad_hook(optimizer_hook)

        optimizer = DummyOptimizer()
    else:
        param_groups = [dict(params=non_galore_params), dict(params=galore_params, **galore_kwargs)]
        optimizer = optim_class(param_groups, **optim_kwargs)

    logger.info("Using GaLore optimizer, may cause hanging at the start of training, wait patiently.")
    return optimizer

def create_custom_optimzer(
    model: "PreTrainedModel",
    dataset: Union["Dataset", "IterableDataset"],
    training_args: "Seq2SeqTrainingArguments",
    **kwargs,
) -> Optional["torch.optim.Optimizer"]:
    finetune_type = kwargs['model_args'].get('finetune_type', 'lora')
    loraplus_lr_ratio = kwargs['model_args'].get('loraplus_lr_ratio', 0)
    use_galore = kwargs['model_args'].get('use_galore', False)
    if finetune_type == 'full' and use_galore:
        return create_galore_optimzer(model,dataset,training_args)
    if finetune_type == 'lora' and loraplus_lr_ratio > 0:
        return create_loraplus_optimizer(model, training_args, **kwargs)
    return None
